import os
import json
from collections import namedtuple

INVALID = '__INVALID__'  # represents invalid user input
HOME_DIR = os.path.expanduser('~')
CONFIG_DIR = os.environ.get('XDG_CONFIG_HOME', os.path.join(HOME_DIR, '.config'))
if not os.path.exists(CONFIG_DIR):
    os.mkdir(CONFIG_DIR)
PUSHY_CONFIG_DIR = os.path.join(CONFIG_DIR, 'pushy-git')
if not os.path.exists(PUSHY_CONFIG_DIR):
    os.mkdir(PUSHY_CONFIG_DIR)

setting = namedtuple('Setting',
                     ['name', 'default',
                      'required', 'convert_to_list', 'hidden'])

DEFAULT_SETTINGS = [
    setting('hostname',
            default='pushy-git',
            required=False,
            convert_to_list=False,
            hidden=False),
    setting('repos',
            default='REQUIRED comma-separated list of repos to check',
            required=True,
            convert_to_list=True,
            hidden=False),
    setting('sleep time',
            default=5,
            required=False,
            convert_to_list=False,
            hidden=False),
    setting('initial pull',
            default=True,
            required=False,
            convert_to_list=False,
            hidden=True),
    setting('update commands',
            default='git pull, git add ., git commit -m,'  # Adds hostname/time
                    'git push -u origin master',
            required=False,
            convert_to_list=True,
            hidden=True),
]


class Config:

    def __init__(self, config_file='pushy_git_settings.json'):
        """
        Reads user settings from the config file, or prompts the user to set
        their config if the file does not exist.

        :param config_file: str, relative path to the config file from HOME_DIR

        If no existing config file is found, the user will be prompted for
        their settings and a config file is created.

        The config file is JSON format.
        """
        self.config_file = os.path.join(PUSHY_CONFIG_DIR, config_file)
        if os.path.exists(self.config_file):
            with open(self.config_file) as config_file:
                user_settings = json.load(config_file)
            self.settings = self.configure_settings(user_settings)
        else:
            self.settings = self.configure_settings({}, prompt_user=True)
            self.write_settings(self.settings)

    def configure_settings(self, user_settings,
                           prompt_user=False, prompt_hidden=False):
        """
        Runs through every setting and, in cases where the user has not
        configured a setting, provides the default value.

        :param user_settings: dictionary, containing settings read from
            the config file

        returns a dictionary containing the settings
        """
        for setting in DEFAULT_SETTINGS:
            if prompt_user:
                value = self.prompt_user(setting, prompt_hidden)
            else:
                value = user_settings.get(setting.name, setting.default)
            if setting.convert_to_list:
                if not isinstance(value, list):
                    value = self.convert_to_list(value)
            else:
                value = self.fix_type(value)
            user_settings[setting.name] = value
        return user_settings

    def prompt_user(self, setting, prompt_hidden=False, value=INVALID):
        """
        Ask the user to configure a setting.

        :param setting: namedtuple, see definition up top
        :prompt_hidden: bool, whether to ask the user to configure a setting
            flagged as hidden

        returns the input provided by the user, or INVALID if the input
        is not acceptable.
        """
        if setting.hidden and not prompt_hidden:
            return setting.default
        while value is INVALID:
            value = input('{} ({}): '.format(setting.name, setting.default))
            if setting.required and not value:
                continue
            elif not value:
                return setting.default
            else:
                return value

    @staticmethod
    def convert_to_list(comma_separated_items):
        """
        Converts a string into a list, handling relative and absolute file
        paths if applicable.

        :param comma_separated_items: str, comma separated list of items to
            convert into a list

        returns list of items
        """
        if not isinstance(comma_separated_items, str):
            raise ValueError('Cannot turn to list: {}'.format(
                comma_separated_items))
        value_split = [item.strip() for item in
                       comma_separated_items.split(',')]
        values_list = []
        for item in value_split:
            if item.startswith('~/'):
                values_list.append(os.path.join(HOME_DIR, item.strip('~/')))
            else:
                values_list.append(item)
        return values_list

    @staticmethod
    def fix_type(value):
        """
        :param value: string containing data provided by the user
        :param convert_to_list: bool

        The value provided will be converted to a list if indicated,
        transformed into an int if possible, or left unchanged,
        and then returned.
        """
        try:
            value = int(value)
        except (ValueError, TypeError):
            pass
        return value

    def write_settings(self, settings):
        with open(self.config_file, 'w') as config_file:
            json.dump(settings, config_file)
