#!/usr/bin/env/python3

import os
import subprocess
import time
import logging
from datetime import datetime


class Pushy:
    """
    git-pushy regularly checks a list of git repos for changes, and issues
    a list of shell commands when a change is detected. By default these
    commands will pull remote changes to the local branch, and push local
    changes to the remote branch.
    """
    def __init__(self, config, log_file):
        """
        :param config: object with a settings dictionary that will be used
            for configuration:

                {'repos': list of file paths for the repos to check,
                 'hostname': str used to format commit messages,
                 'sleep time': int for number of seconds to sleep between
                    checks,
                 'inital pull': bool, if True the repos being tracked will
                    have remote changes pulled before the regular scanning
                    begins,
                 'update commands': list of shell comands to run when a change
                    has been detected in a git repo}

        :param log_file: file path to the log file

        A Config object is passed instead of just a dictionary of settings so
        that it is easy to call on the methods for getting/setting
        configuration from the context object used by the Click library.
        """
        for setting, value in config.settings.items():
            self.__setattr__(setting.replace(' ', '_'), value)
        self.config = config
        logging.basicConfig(filename=log_file, level=logging.DEBUG)
        self.quit = False

    def __call__(self):
        """
        Runs a loop forever, calling the main method.
        """
        if self.initial_pull:
            time.sleep(180)  # A few minutes to connect to internet
            for repo_path in self.repos:
                os.chdir(repo_path)
                self.git_command('git pull')
        while not self.quit:
            self.main()

    def main(self):
        """
        Sleeps between checks, then checks every tracked repo for changes
        and runs the update commands if any are found.

        This method changes the working directory to every repo's file path
        in order to run the git shell commands. I just...thought you should
        know.
        """
        time.sleep(self.sleep_time)
        for repo_path in self.repos:
            os.chdir(repo_path)
            git_status = self.git_command('git status')
            if self.untracked_changes(git_status):
                self.update()

    def git_command(self, command):
        """
        Calls the git commands from the shell.

        :param command: str of a git shell command

        returns: str with stdout of shell command
        """
        with subprocess.Popen(command, stdout=subprocess.PIPE,
                              shell=True) as proc:
            return proc.stdout.read().decode()

    def untracked_changes(self, git_status):
        """
        :param git_status_stdout: str, stdout from the git status shell command

        returns True if it finds untracked changes
        """
        return 'Changes not staged' in git_status or \
               'Untracked files' in git_status or \
               'branch is ahead' in git_status or \
               'Changes to be committed' in git_status

    def update(self):
        """
        Runs every command in self.git_update_commands (see HIDDEN_SETTINGS
        for the defaults).
        Formats the command string with the commit message when calling it.
        """
        for command in self.update_commands:
            if 'commit' in command:
                command += " '{}: {}'".format(self.hostname, datetime.utcnow())
            print(command)
            command_output = self.git_command(command)
            logging.info(command_output)
